/*
 * This file is part of BukkitHTTPD.
 * Copyright (C) 2011 "codename_B" Ben Smith <BenBenBen001@googlemail.com>
 * Copyright (C) 2011 Thomas "V10lator" Rohloff <v10lator@myway.de>
 * 
 *
 * BukkitHTTPD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BukkitHTTPD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BukkitHTTPD.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.V10lator.BukkitHTTPD;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.Map.Entry;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.logging.Logger;

import org.bukkit.Server;
import org.bukkit.configuration.InvalidConfigurationException;
import org.bukkit.configuration.file.YamlConfiguration;
import org.bukkit.event.Event;
import org.bukkit.plugin.PluginDescriptionFile;
import org.bukkit.plugin.PluginManager;
import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.scheduler.BukkitScheduler;

/**
 * 
 * @author codename_B
 * @author V10lator
 *
 */
public final class BukkitHTTPD extends JavaPlugin {

	/**
	 * The main configuration file.
	 */
	private final File f = new File("plugins/BukkitHTTPD.yml");
	final YamlConfiguration config = new YamlConfiguration();
	
	/**
	 * The save file.
	 */
	private final File sf = new File("plugins/BukkitHTTPD.sav");
	final YamlConfiguration save = new YamlConfiguration();
	/**
	 * The logging utility (used for error logging).
	 */
	private Logger log;
	
	/**
	 * The NanoHTTPD session
	 */
	private NanoHTTPD httpd;
	
	/**
	 * The PHP executable(full path)
	 */
	AtomicBoolean phpEnabled = new AtomicBoolean(false);
	String php;
	final String phpCode = " -d html_errors=1 -d implicit_flush=0 -d max_execution_time=30";
	
	/**
	 * RAM and buffers between BukkitHTTPD and NanoHTTPD/V10parser.
	 */
	final RAM ram = new RAM();
	String[] chatBuffer = new String[20];
	int itimer;
	
	/**
	 * Listeners
	 */
	BukkitHTTPDPlayerListener pL = new BukkitHTTPDPlayerListener(this);
	
    /**
	 * The port and the directory of the HTTP daemon server.
	 */
	int httpdPort;
	File wwwDir;
	
	boolean createBanner;
	int bannerPID;
	
	public void onDisable()
	{
	  if(httpd != null)
	    httpd.stop();
	  getServer().getScheduler().cancelTasks(this);
	  saveConfig();
	  log.info("["+getDescription().getName()+"] disabled!");
	}

	public void onEnable()
	{
	  Server s = getServer();
	  log = s.getLogger();
	  loadConfig();
	  saveConfig();
	  wwwDir.mkdirs();
	  checkFiles();
	  BukkitScheduler bs = s.getScheduler();
	  bs.scheduleSyncRepeatingTask(this, new RAMFiller(this), 0L, 3L);
	  try
	  {
		httpd = new NanoHTTPD(this, httpdPort, wwwDir, php);
	  }
	  catch (IOException e)
	  {
		log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		s.getPluginManager().disablePlugin(this);
		return;
	  }
	  if(createBanner)
	    bannerPID = bs.scheduleAsyncRepeatingTask(this, new ImageCreator(this), 40L, itimer * 20);
	  PluginManager pm = s.getPluginManager();
	  pm.registerEvent(Event.Type.PLAYER_CHAT, pL, Event.Priority.Monitor, this);
	  pm.registerEvent(Event.Type.PLAYER_JOIN, pL, Event.Priority.Monitor, this);
	  pm.registerEvent(Event.Type.PLAYER_QUIT, pL, Event.Priority.Monitor, this);
	  pm.registerEvent(Event.Type.PLAYER_KICK, pL, Event.Priority.Monitor, this);
	  getCommand("httpd").setExecutor(new BukkitHTTPDCommandParser(this));
	  PluginDescriptionFile pdf = getDescription();
	  log.info("["+pdf.getName()+"] v"+pdf.getVersion()+" enabled!");
	}
	
	public void loadConfig()
	{
	  String www;
	  try
	  {
		config.load(f);
	  }
	  catch(FileNotFoundException e)
	  {
		try
		{
		  f.createNewFile();
		}
		catch
		(IOException e1)
		{
		  log.info("["+getDescription().getName()+"] error!");
		  e.printStackTrace();
		  getServer().getPluginManager().disablePlugin(this);
		  return;
		}
	  }
	  catch(IOException e)
	  {
		log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		getServer().getPluginManager().disablePlugin(this);
		return;
	  }
	  catch(InvalidConfigurationException e)
	  {
		log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		getServer().getPluginManager().disablePlugin(this);
		return;
	  }
	  try
	  {
		save.load(sf);
	  }
	  catch(FileNotFoundException e)
	  {
		try
		{
		  sf.createNewFile();
		}
		catch
		(IOException e1)
		{
		  log.info("["+getDescription().getName()+"] error!");
		  e.printStackTrace();
		  getServer().getPluginManager().disablePlugin(this);
		  return;
		}
	  }
	  catch(IOException e)
	  {
		log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		getServer().getPluginManager().disablePlugin(this);
		return;
	  }
	  catch(InvalidConfigurationException e)
	  {
		log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		getServer().getPluginManager().disablePlugin(this);
		return;
	  }
	  
	  while(!ram.lock.compareAndSet(false, true))
		continue;
	  
	  httpdPort = config.getInt("Port", 8107);
	  www = config.getString("Directory", "plugins/BukkitHTTPD/");
	  createBanner = config.getBoolean("Banner", true);
	  ram.hidePort = config.getBoolean("HidePort", false);
	  ram.serverName = config.getString("ServerName", "My Awesome Server");
	  ram.color = config.getString("Color", "000000");
	  if(!www.endsWith("/"))
		www += "/";
	  wwwDir = new File(www);
	  itimer = config.getInt("BannerRefresh", 30);
	  ram.error[0] = config.getString("Error.404", "error/404.html");
	  ram.error[1] = config.getString("Error.403", "error/403.html");
	  
	  phpEnabled.getAndSet(false);
	  php = config.getString("PHP.exe", "/usr/bin/php");
	  phpEnabled.set(config.getBoolean("PHP.Enabled", false));
	  
	  ram.users.clear();
	  for(String sp: save.getKeys(false))
	  {
		WebUser user = new WebUser();
		user.pwd = save.getString(sp+".hash");
		user.isAdmin = save.getBoolean(sp+".isAdmin", false);
		ram.users.put(sp, user);
	  }
	  ram.lock.set(false);
	}
	
	public void saveConfig()
	{
	  config.set("HidePort", ram.hidePort);
	  config.set("Port", httpdPort);
	  config.set("Directory", wwwDir.toString());
	  config.set("ServerName", null);
	  config.set("Color", ram.color);
	  config.set("MinecraftBans", ram.minecraftBans);
	  config.set("BannerRefresh", itimer);
	  config.set("error.404", ram.error[0]);
	  config.set("error.403", ram.error[1]);
	  config.set("Banner", createBanner);
	  
	  boolean pe = phpEnabled.getAndSet(false);
	  config.set("PHP.exe", php);
	  config.set("PHP.Enabled", pe);
	  phpEnabled.set(pe);
	  
	  //TODO: Test
	  for(String sk: save.getKeys(false))
		save.set(sk, null);
	  
	  for(Entry<String, WebUser> e: ram.users.entrySet())
	  {
		String k = e.getKey();
		WebUser user = e.getValue();
		save.set(k+".hash", user.pwd);
		save.set(k+".isAdmin", user.isAdmin);
	  }
	  ram.lock.set(false);
	  try
	  {
	    config.save(f);
	    save.save(sf);
	  }
	  catch(IOException e)
	  {
		log.info("["+getDescription().getName()+"] could not save!");
	    e.printStackTrace();
	  }
	}
	
	public void reloadConfig()
	{
	  loadConfig();
	  httpd.stop();
	  BukkitScheduler bs = getServer().getScheduler();
	  bs.cancelTasks(this);
	  saveConfig();
	  wwwDir.mkdirs();
	  checkFiles();
	  bs.scheduleSyncRepeatingTask(this, new RAMFiller(this), 0L, 3L);
	  try
	  {
		httpd = new NanoHTTPD(this, httpdPort, wwwDir, php);
	  }
	  catch (IOException e)
	  {
		log.info("["+getDescription().getName()+"] error!");
		e.printStackTrace();
		getServer().getPluginManager().disablePlugin(this);
		return;
	  }
	  bs.scheduleAsyncRepeatingTask(this, new ImageCreator(this), 0L, itimer * 20);
	}
	
	public void checkFiles()
	{
	  String prefix = "["+getDescription().getName()+"] ";
	  if(!wwwDir.exists())
	  {
		if(wwwDir.mkdirs())
	      log.info(prefix+"www folder created.");
	    else
	    {
	      log.info(prefix+"can't create www folder!");
	      return;
	    }
	  }
	  
	  try
	  {
		JarFile jar = new JarFile(getFile());
		Enumeration<JarEntry> entries = jar.entries();
		String pkg = jar.getEntry(getClass().getPackage().getName().replace(".", "/")+"/www/").getName();
		File f;
		while(entries.hasMoreElements())
		{
		  JarEntry entry = entries.nextElement();
		  String en = entry.getName();
		  if(!en.startsWith(pkg))
			continue;
		  if(entry.isDirectory())
			new File(wwwDir, en.substring(pkg.length())).mkdirs();
		  else
		  {
			f = new File(wwwDir, en.substring(pkg.length()));
			if(!f.canRead())
			{
			  try
			  {
				if(!f.exists())
				{
				  f.createNewFile();
				  InputStream is = jar.getInputStream(entry);
				  FileOutputStream os = new FileOutputStream(f);
				  int i;
				  while((i = is.read()) > -1)
				    os.write(i);
				  is.close();
				  os.close();
				}
			  }
			  catch (IOException e)
		      {
				log.info(prefix+"Can't create default file!");
			    e.printStackTrace();
		      }
			}
		  }
		}
	  }
	  catch(IOException e)
	  {
		e.printStackTrace();
		log.info(prefix+"Can't create default file!");
		e.printStackTrace();
	  }
}
}