/*
 * This file is part of BukkitHTTPD.
 * Copyright (C) 2011 "codename_B" Ben Smith <BenBenBen001@googlemail.com>
 * Copyright (C) 2011 Thomas "V10lator" Rohloff <v10lator@myway.de>
 * 
 *
 * BukkitHTTPD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BukkitHTTPD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BukkitHTTPD.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.V10lator.BukkitHTTPD;

import java.io.File;
import java.nio.CharBuffer;
import java.nio.charset.CharacterCodingException;
import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Iterator;

import org.bukkit.ChatColor;
import org.bukkit.Server;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;

/**
 * 
 * @author V10lator
 * 
 */
public class BukkitHTTPDCommandParser implements CommandExecutor
{
  private final BukkitHTTPD plugin;
  
  public BukkitHTTPDCommandParser(BukkitHTTPD plugin)
  {
	this.plugin = plugin;
  }
  
  public boolean onCommand(CommandSender sender, Command arg1, String arg2,
			String[] args)
  {
	boolean pw = (args.length > 1 && args[0].equalsIgnoreCase("password"));
	if(!sender.hasPermission("bukkithttpd.admin") || pw)
	{
	  if(!sender.hasPermission("bukkithttpd.user") || !(sender instanceof Player))
	    return true;
	  if(!pw)
	    sendHelp(sender);
	  else
	  {
		String passwd;
		if(args.length > 2)
		{
		  StringBuilder sb = new StringBuilder(args[1]);
		  for(int i = 2; i < args.length; i++)
		  {
			sb.append(" ");
			sb.append(args[i]);
		  }
		  passwd = sb.toString();
		}
		else
		  passwd = args[1];
		try
		{
		  MessageDigest md = MessageDigest.getInstance("SHA");
		  md.update(Charset.forName("UTF-8").newEncoder().encode(CharBuffer.wrap(passwd)));
		  byte[] sha = md.digest();
		  StringBuffer hex = new StringBuffer();
		  for(int i=0; i < sha.length; i++)
		    hex.append(Integer.toHexString(0xFF & sha[i]));
		  
		  while(plugin.ram.lock.compareAndSet(false, true))
			continue;
		  String sn = ((Player)sender).getName();
		  WebUser wu;
		  if(plugin.ram.users.containsKey(sn))
			wu = plugin.ram.users.get(sn);
		  else
		  {
			wu = new WebUser();
			plugin.ram.users.put(sn,  wu);
		  }
		  wu.pwd = hex.toString();
		  wu.isOnline = false;
		  plugin.ram.lock.set(false);
		  plugin.saveConfig();
		  sender.sendMessage(ChatColor.GREEN+"Password set.");
		}
		catch(CharacterCodingException e)
		{
		  sender.sendMessage(ChatColor.RED+"Internal error!");
		  e.printStackTrace();
		}
		catch(NoSuchAlgorithmException e)
		{
		  sender.sendMessage(ChatColor.RED+"Internal error!");
		  e.printStackTrace();
		}
	  }
	  return true;
	}
	
	while(plugin.ram.lock.compareAndSet(false, true))
	  continue;
	
	boolean save = false;
	boolean reload = false;
	
	if(args.length < 1 || args[0].equalsIgnoreCase("help"))
	  sendHelp(sender);
	else if(args[0].equalsIgnoreCase("port"))
	{
	  if(args.length < 2)
	  {
		sender.sendMessage(ChatColor.GREEN+"Current server port:");
		sender.sendMessage(ChatColor.YELLOW+String.valueOf(plugin.httpdPort)+" ("+ChatColor.GOLD+plugin.ram.hidePort+ChatColor.YELLOW+")");
	  }
	  else
	  {
		try
		{
		  if(args[1].equalsIgnoreCase("hide"))
		  {
			if(plugin.ram.hidePort)
			{
			  plugin.ram.hidePort = false;
			  sender.sendMessage(ChatColor.YELLOW+"Hiding port.");
			}
			else
			{
			  plugin.ram.hidePort = true;
			  sender.sendMessage(ChatColor.GREEN+"Showing port.");
			}
		  }
		  else
		  {
		    plugin.httpdPort = Integer.parseInt(args[1]);
		    sender.sendMessage(ChatColor.YELLOW+"New server port:");
		    sender.sendMessage(ChatColor.GREEN+args[1]);
		  }
		  save = true;
		}
		catch(NumberFormatException e)
		{
		  sender.sendMessage(ChatColor.RED+"Invalid port: \""+ChatColor.GOLD+args[1]+ChatColor.RED+"\"");
		}
	  }
	}
	else if(args[0].equalsIgnoreCase("color"))
	{
	  if(args.length < 2)
	  {
		sender.sendMessage(ChatColor.GREEN+"Current color:");
		sender.sendMessage(ChatColor.YELLOW+plugin.ram.color);
	  }
	  else
	  {
		try
		{
		  Integer.parseInt(args[1], 16);
		  if(args[1].length() != 6)
			throw new NumberFormatException();
		  plugin.ram.color = args[1];
		  save = true;
		  sender.sendMessage(ChatColor.YELLOW+"New color:");
		  sender.sendMessage(ChatColor.GREEN+plugin.ram.color);
		}
		catch(NumberFormatException e)
		{
		  sender.sendMessage(ChatColor.RED+"Invalid color: \""+ChatColor.GOLD+args[1]+ChatColor.RED+"\"");
		}
	  }
	}
	else if(args[0].equalsIgnoreCase("plugins"))
	{
	  if(plugin.ram.showPlugins)
	  {
		plugin.ram.showPlugins = false;
		sender.sendMessage(ChatColor.YELLOW+"Hiding plugins.");
	  }
	  else
	  {
		plugin.ram.showPlugins = true;
		sender.sendMessage(ChatColor.YELLOW+"Showing plugins.");
	  }
	  save = true;
	}
	else if(args[0].equalsIgnoreCase("refresh"))
	{
	  try
	  {
		int r = Integer.parseInt(args[1]);
		if(r < 1)
		  sender.sendMessage(ChatColor.RED+"Invalid refresh rate: \""+ChatColor.GOLD+args[1]+ChatColor.RED+"\"");
		else
		{
		  plugin.itimer = r;
		  sender.sendMessage(ChatColor.GREEN+"Refresh rate set!");
		  save = true;
		  reload = true;
		}
	  }
	  catch(NumberFormatException e)
	  {
		sender.sendMessage(ChatColor.RED+"Invalid refresh rate: \""+ChatColor.GOLD+args[1]+ChatColor.RED+"\"");
	  }
	}
	else if(args[0].equalsIgnoreCase("dir"))
	{
	  if(args.length < 2)
	  {
		sender.sendMessage(ChatColor.GREEN+"Current directory:");
		sender.sendMessage(ChatColor.YELLOW+plugin.wwwDir.toString());
	  }
	  else
	  {
		StringBuilder sb = new StringBuilder(args[1]);
		for(int i = 2; i < args.length; i++)
		{
		  sb.append(" ");
		  sb.append(args[i]);
		}
		plugin.wwwDir = new File(sb.toString());
		save = true;
		reload = true;
		sender.sendMessage(ChatColor.YELLOW+"New directory:");
		sender.sendMessage(ChatColor.GREEN+plugin.wwwDir.toString());
	  }
	}
	else if(args[0].equalsIgnoreCase("bans"))
	{
	  if(plugin.ram.minecraftBans)
	  {
		plugin.ram.minecraftBans = false;
		sender.sendMessage(ChatColor.YELLOW+"Ignoring minecraft bans.");
	  }
	  else
	  {
		plugin.ram.minecraftBans = true;
		sender.sendMessage(ChatColor.GREEN+"Using minecraft bans.");
	  }
	}
	else if(args[0].equalsIgnoreCase("admin") && args.length > 1)
	{
	  if(!plugin.ram.users.containsKey(args[1]))
		sender.sendMessage(ChatColor.RED+args[1]+" not found!");
	  else
	  {
		WebUser user = plugin.ram.users.get(args[1]);
		if(user.isAdmin)
		{
		  user.isAdmin = false;
		  sender.sendMessage(ChatColor.YELLOW+args[1]+" is no longer web admin.");
		}
		else
		{
		  user.isAdmin = true;
		  sender.sendMessage(ChatColor.YELLOW+args[1]+" is web admin now.");
		}
		save = true;
	  }
	}
	else if(args[0].equalsIgnoreCase("error"))
	{
	  if(args.length < 2)
		sender.sendMessage(ChatColor.RED+"No error number given!");
	  else if(args.length == 2)
	  {
		String f = null;
		if(args[1].equals("404"))
		  f = plugin.ram.error[0];
		else if(args[1].equals("403"))
		  f = plugin.ram.error[1];
		if(f != null)
		{
		  sender.sendMessage(ChatColor.YELLOW+"Current "+args[0]+" error file:");
		  sender.sendMessage(ChatColor.GREEN+f);
		}
		else
		  sender.sendMessage(ChatColor.RED+"Wrong error number: "+args[1]);
	  }
	  else
	  {
		StringBuilder sb = new StringBuilder(args[2]);
		for(int i = 3; i < args.length; i++)
		{
		  sb.append(" ");
		  sb.append(args[i]);
		}
		String f = sb.toString();
		boolean done = true;
		if(args[1].equals("404"))
		  plugin.ram.error[0] = f;
		else if(args[1].equals("403"))
		  plugin.ram.error[1] = f;
		else
		{
		  sender.sendMessage(ChatColor.RED+"Wrong error number: "+args[1]);
		  done = false;
		}
		if(done)
		{
		  sender.sendMessage(ChatColor.YELLOW+"New "+args[1]+" error file:");
		  sender.sendMessage(ChatColor.GREEN+f);
		  save = true;
		}
	  }
	}
	else if(args[0].equalsIgnoreCase("php"))
	{
	  if(args.length < 2)
	  {
		boolean p = plugin.phpEnabled.getAndSet(false);
		if(p)
		  sender.sendMessage(ChatColor.YELLOW+"PHP disabled.");
		else
		{
		  plugin.phpEnabled.getAndSet(true);
		  sender.sendMessage(ChatColor.YELLOW+"PHP enabled.");
		}
	  }
	  else
	  {
		StringBuilder sb = new StringBuilder(args[1]);
		for(int i = 2; i < args.length; i++)
		{
		  sb.append(" ");
		  sb.append(args[i]);
		}
		String p = sb.toString();
		boolean pe = plugin.phpEnabled.getAndSet(false);
		plugin.php = p;
		plugin.phpEnabled.set(pe);
		sender.sendMessage(ChatColor.YELLOW+"New php binary:");
		sender.sendMessage(ChatColor.GREEN+p);
	  }
	  save = true;
	}
	else if(args[0].equalsIgnoreCase("reload"))
	  reload = true;
	//TODO: Test.
	else if(args[0].equalsIgnoreCase("rmpwd") && args.length > 1)
	{
	  if(plugin.ram.users.containsKey(args[1]))
	  {
		plugin.ram.users.remove(args[1]);
		save = true;
	  }
	  else
		sender.sendMessage(ChatColor.RED+args[1]+" not found!");
	}
	else if(args[0].equalsIgnoreCase("banner"))
	{
	  if(plugin.createBanner)
	  {
		plugin.getServer().getScheduler().cancelTask(plugin.bannerPID);
		plugin.createBanner = false;
		File b = new File(plugin.wwwDir, "banner.png");
		if(b.exists())
		  b.delete();
		sender.sendMessage(ChatColor.GOLD+"Banner disabled!");
	  }
	  else
	  {
		plugin.bannerPID = plugin.getServer().getScheduler().scheduleAsyncRepeatingTask(plugin, new ImageCreator(plugin), 40L, plugin.itimer * 20);
		plugin.createBanner = true;
		sender.sendMessage(ChatColor.GREEN+"Banner enabled!");
	  }
	  save = true;
	}
	else
	  sendHelp(sender);
	
	plugin.ram.lock.set(false);
	if(save)
	{
	  plugin.saveConfig();
	  if(reload)
	    plugin.reloadConfig();
	  sender.sendMessage(ChatColor.GREEN+"Config saved!");
	}
	else if(reload)
	{
	  plugin.reloadConfig();
	  sender.sendMessage(ChatColor.GREEN+"Config reloaded!");
	}
	return true;
  }
  
  private void sendHelp(CommandSender sender)
  {
	ArrayList<String> message = new ArrayList<String>();
	boolean admin = sender.hasPermission("bukkithttpd.admin");
	boolean user = sender.hasPermission("bukkithttpd.user");
	boolean player = sender instanceof Player;
	String prefix;
	  if(player)
		prefix = "/";
	  else
		prefix = "";
	if(admin)
	{
	  message.add(prefix+"httpd port");
	  message.add("    Show/set the server port.");
	  message.add(prefix+"httpd port hide");
	  message.add("    Toggle the port hiding switch.");
	  message.add(prefix+"httpd banner");
	  message.add("    Toggle the banner switch.");
	  message.add(prefix+"httpd color");
	  message.add("    Show/set the banner color.");
	  message.add(prefix+"httpd refresh");
	  message.add("    Set the banner refresh rate.");
	  message.add(prefix+"httpd dir");
	  message.add("    Show/set the www folder.");
	  message.add(prefix+"httpd php");
	  message.add("    Change php settings.");
	  message.add(prefix+"httpd bans");
	  message.add("    Toggle the minecraft bans feature.");
	  message.add(prefix+"httpd error");
	  message.add("    Show/set the 404 and 403 error pages.");
	  message.add(prefix+"httpd reload");
	  message.add("    Reload the configuration from disk.");
	  //TODO: Test.
	  message.add(prefix+"httpd rmpwd <User>");
	  message.add("    Delete the BukkitHTTPD password of user <User>.");
	}
	if(user && player)
	{
	  message.add("/httpd password <password>");
	  message.add("    Set your BukkitHTTPD password.");
	}
	if(user || admin)
	{
	  message.add(prefix+"httpd help");
	  message.add("    Show this help.");
	}
	
	int s = message.size();
	if(s < 1)
	  return;
	Iterator<String> iter = message.iterator();
	if(s < 21)
	  while(iter.hasNext())
		sender.sendMessage(iter.next());
	else
	  plugin.getServer().getScheduler().scheduleSyncDelayedTask(plugin, new HelpReader(sender, iter));
  }
  
  private class HelpReader implements Runnable
  {
	private final String player;
	private final Iterator<String> messages;
	
	private HelpReader(CommandSender sender, Iterator<String> messages)
	{
	  this.messages = messages;
	  if(sender instanceof Player)
		player = ((Player)sender).getName();
	  else
		player = null;
	}
	
	public void run()
	{
	  Server s = plugin.getServer();
	  CommandSender sender;
	  if(player != null)
	  {
		Player p = s.getPlayerExact(player);
		if(p == null || !p.isOnline())
		  return;
		sender = (CommandSender)p;
	  }
	  else
		sender = s.getConsoleSender();
	  
	  int c = 0;
	  while(messages.hasNext())
	  {
		sender.sendMessage(messages.next());
		c++;
		if(c > 1)
		  break;
	  }
	  if(messages.hasNext())
		s.getScheduler().scheduleSyncDelayedTask(plugin, this, 40);
	}
  }
}
